import 'package:extensions_pack/extensions_pack.dart';
import 'package:flutter/material.dart';
import 'package:widgets_pack/helpers/helpers.dart';
import 'package:widgets_pack/widgets/widgets.dart';

part 'table_config_consumer.dart';
part 'table_context.extension.dart';
part 'table_controller.dart';
part 'table_filter.dart';

class AppTable<M extends Object> extends StatefulWidget {
  final TableController<M> controller;
  final List<TableAction<M>> actions;
  final List<TableColumn<M>> columns;
  final List<Widget> filters;
  final void Function(M)? onRowTap;
  final void Function(M)? onRowDoubleTap;
  final AppButtonConfig? headerAction;

  const AppTable({
    required this.controller,
    required this.columns,
    this.actions = const [],
    this.filters = const [],
    this.onRowTap,
    this.onRowDoubleTap,
    this.headerAction,
    super.key,
  });

  @override
  State<AppTable<M>> createState() => _AppTableState<M>();

  static _AppTableState<M> of<M extends Object>(BuildContext context) {
    return context.findAncestorStateOfType<_AppTableState<M>>()!;
  }
}

class _AppTableState<M extends Object> extends State<AppTable<M>> {
  late TableDataSource<M> _dataSource;

  @override
  void initState() {
    super.initState();

    _dataSource = TableDataSource<M>(
      loadFn: widget.controller.loadFn,
      columns: widget.columns,
      actions: widget.actions,
    );

    widget.controller.dataSource = _dataSource;
  }

  @override
  void didUpdateWidget(covariant AppTable<M> oldWidget) {
    super.didUpdateWidget(oldWidget);

    if (oldWidget.columns != widget.columns) {
      _dataSource.setColumns(widget.columns);
    }

    if (oldWidget.actions != widget.actions) {
      _dataSource.setActionsType(widget.actions);
    }
  }

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
      builder: (context, constraints) {
        return _DatasourceConfigConsumer<M, List<TableColumn<M>>>(
          selector: (source) => source.columns,
          builder: (context, columns) {
            final fixedColumns = columns.where((c) => c.fixed).toList();
            final normalColumns = columns.where((c) => !c.fixed).toList();
            final fixedColumnsWidth = fixedColumns.fold<double>(
              0,
              (previousValue, element) => previousValue + element.width,
            );

            return Column(
              mainAxisSize: MainAxisSize.min,
              children: [
                _DatasourceConfigConsumer<M, int>(
                  selector: (source) => source.selectedItems.length,
                  builder: (context, length) {
                    return AnimatedSwitcher(
                      duration: const Duration(milliseconds: 300),
                      child: length == 0
                          ? _AppTableFilterRow(
                              headerAction: widget.headerAction,
                              filters: widget.filters,
                            )
                          : AppTableActionsRow(
                              items: _dataSource.selectedItems,
                              onClearAll: _dataSource.clearSelection,
                              actions: widget.actions,
                            ),
                    );
                  },
                ),
                Flexible(
                  child: Row(
                    crossAxisAlignment: CrossAxisAlignment.start,
                    children: [
                      if (fixedColumns.isNotEmpty)
                        SizedBox(
                          width: fixedColumnsWidth,
                          child: _Table<M>(
                            key: const Key('fixed_columns_table'),
                            columns: fixedColumns,
                            dataSource: _dataSource,
                            rowBorder: Border(
                              right: BorderSide(
                                width: 0.5,
                                color: Theme.of(context).dividerColor,
                              ),
                            ),
                            showLoader: false,
                            onRowTap: widget.onRowTap,
                            onRowDoubleTap: widget.onRowDoubleTap,
                          ),
                        ),
                      Expanded(
                        child: SingleChildScrollView(
                          physics: const ClampingScrollPhysics(),
                          scrollDirection: Axis.horizontal,
                          child: SizedBox(
                            width: normalColumns
                                .fold<double>(0, (previousValue, element) => previousValue + element.width)
                                .clamp(constraints.minWidth - fixedColumnsWidth, double.maxFinite),
                            child: _Table<M>(
                              key: const Key('normal_columns_table'),
                              columns: normalColumns,
                              dataSource: _dataSource,
                              onRowTap: widget.onRowTap,
                              onRowDoubleTap: widget.onRowDoubleTap,
                            ),
                          ),
                        ),
                      ),
                    ],
                  ),
                ),
                ConstrainedBox(
                  constraints: const BoxConstraints(minHeight: 52),
                  child: _DatasourceConfigConsumer<M, TableDatasourceConfig<M>>(
                    selector: (source) => source.config,
                    builder: (context, config) {
                      final firstIndex = config.currentPage * config.pageSize + 1;

                      return Row(
                        mainAxisAlignment: MainAxisAlignment.end,
                        children: [
                          Text(
                            '$firstIndex - ${firstIndex + (config.pages[config.currentPage]?.length ?? 1) - 1}',
                          ),
                          const Spacing(mainAxisExtent: kXSSize),
                          IconButton(
                            onPressed: config.canGoToPreviousPage
                                ? () {
                                    _dataSource.setPage(0);
                                  }
                                : null,
                            icon: const Icon(Icons.skip_previous),
                          ),
                          IconButton(
                            onPressed: config.canGoToPreviousPage
                                ? () {
                                    _dataSource.setPage(config.currentPage - 1);
                                  }
                                : null,
                            icon: const Icon(Icons.chevron_left),
                          ),
                          IconButton(
                            onPressed: config.canGoToNextPage
                                ? () {
                                    _dataSource.setPage(config.currentPage + 1);
                                  }
                                : null,
                            icon: const Icon(Icons.chevron_right),
                          ),
                          IconButton(
                            onPressed: config.canGoToNextPage
                                ? () {
                                    _dataSource.setPage(config.lastAvailablePage);
                                  }
                                : null,
                            icon: const Icon(Icons.skip_next),
                          ),
                        ],
                      );
                    },
                  ),
                ),
              ],
            );
          },
        );
      },
    );
  }
}

class _Table<M extends Object> extends StatelessWidget {
  final List<TableColumn<M>> columns;
  final TableDataSource<M> dataSource;
  final bool showLoader;
  final Border rowBorder;
  final void Function(M)? onRowTap;
  final void Function(M)? onRowDoubleTap;

  const _Table({
    required this.columns,
    required this.dataSource,
    this.showLoader = true,
    this.rowBorder = const Border(),
    this.onRowTap,
    this.onRowDoubleTap,
    super.key,
  });

  @override
  Widget build(BuildContext context) {
    return _DatasourceConfigConsumer<M, int>(
      selector: (source) => source.pageSize,
      builder: (context, pageSize) {
        return ListView.separated(
          itemCount: pageSize + 1,
          shrinkWrap: true,
          physics: const NeverScrollableScrollPhysics(),
          separatorBuilder: (context, index) => const Divider(height: 0),
          itemBuilder: (context, index) {
            final isFirst = index == 0;
            final itemIndex = index - 1;

            return Column(
              mainAxisSize: MainAxisSize.min,
              children: [
                if (isFirst)
                  Column(
                    mainAxisSize: MainAxisSize.min,
                    children: [
                      ConstrainedBox(
                        constraints: const BoxConstraints(minHeight: 52),
                        child: Row(
                          children: [
                            for (final column in columns) column.labelBuilder(),
                          ],
                        ),
                      ),
                      _DatasourceConfigConsumer<M, bool>(
                        selector: (source) => source.config.loading,
                        builder: (context, loading) {
                          if (loading && showLoader) {
                            return const LinearProgressIndicator();
                          }

                          return Divider(
                            thickness: loading ? 2 : 0,
                            height: loading ? 4 : 0,
                          );
                        },
                      ),
                    ],
                  )
                else
                  ConstrainedBox(
                    constraints: const BoxConstraints(minHeight: 52),
                    child: _DatasourceConfigConsumer<M, (M?, bool)>(
                      selector: (source) => source.currentItemAt(itemIndex),
                      builder: (context, model) {
                        if (model.$1 == null) {
                          return const SizedBox();
                        }

                        return InkWell(
                          onTap: () {
                            onRowTap?.call(model.$1!);
                            dataSource.addOrRemoveItem(model.$1!);
                          },
                          onDoubleTap: onRowDoubleTap != null
                              ? () {
                                  onRowDoubleTap?.call(model.$1!);
                                }
                              : null,
                          child: DecoratedBox(
                            decoration: BoxDecoration(
                              border: rowBorder,
                              color: model.$2 ? context.colorTheme.primary.withOpacity(0.08) : null,
                            ),
                            child: Row(
                              key: ValueKey((index, model.$1)),
                              children: [
                                for (final column in columns) column.contentBuilder(model.$1!),
                              ],
                            ),
                          ),
                        );
                      },
                    ),
                  ),
              ],
            );
          },
        );
      },
    );
  }
}
